#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 15 22:52:11 2023

@author: seoxubuntu
"""

from setuptools import setup, find_packages
setup(name="sew",
      version="0.0.4",
      packages=find_packages("."))

